package minischeme

import java.io._

/**
 * Scanner for minischeme.
 *
 * @author Michel Schinz
 */
class Scanner(inputStreams: List[(String,InputStream)]) {
  private var nextCharAsInt : Int = _
  private var line, column : Int = 1
  def nextChar: Char = nextCharAsInt.asInstanceOf[Char]

  private var inputStreamReaders: List[(String,InputStreamReader)] =
    inputStreams map { case (n,s) => (n, new InputStreamReader(s)) }

  private def currentPos: Position = inputStreamReaders match {
    case (file, _) :: _ => new Position(file, line, column)
    case _ => new Position("<none>", line, column)
  }

  private def skipChar(): Unit = {
    nextCharAsInt = inputStreamReaders.head._2.read()
    if (nextCharAsInt == -1) {
      line = 1
      column = 1
      inputStreamReaders = inputStreamReaders.tail
      if (!inputStreamReaders.isEmpty) skipChar()
    } else if (nextCharAsInt.asInstanceOf[Char] == '\n') {
      line = line + 1
      column = 1
    } else
      column = column + 1
  }
  skipChar()

  def nextToken(): Token = {
    val l = line
    val c = column
    val token = readNextToken()
    token.pos = currentPos
    token
  }

  private def readNextToken(): Token = {
    if (nextCharAsInt == -1)
      EOF
    else nextChar match {
      case c if (Character.isWhitespace(c)) =>
        while (Character.isWhitespace(nextChar)) skipChar()
        readNextToken()
      case ';' =>
        while (nextChar != '\n') skipChar()
        readNextToken()
      case c if (isNumberCharacter(c)) =>
        NumberToken(Integer.parseInt(readStringWhile(isNumberCharacter)))
      case c if (isIdentStartCharacter(c)) =>
        IdentToken(readStringWhile(isIdentComponentCharacter))
      case '"' =>
        skipChar()
        val res = StringToken(readStringWhile(isStringComponentCharacter))
        skipExpectedChar('"')
        res
      case '#' =>
        skipChar()
        skipExpectedChar('\\')
        val res = NumberToken(nextCharAsInt)
        skipChar()
        res
      case '(' =>
        skipChar()
        LeftParen
      case ')' =>
        skipChar()
        RightParen
      case other =>
        Report.error(currentPos, "Unknown character '" + other + "'")
    }
  }

  private def skipExpectedChar(c: Char): Unit =
    if (nextChar == c)
      skipChar()
    else
      Report.error(currentPos, "expected " + c + ", found " + nextChar)

  private def inCharSet(set: String)(c: Char): Boolean =
    (set indexOf Character.toLowerCase(c)) != -1
  private val isNumberCharacter =
    inCharSet("0123456789")_
  private val isIdentStartCharacter =
    inCharSet("abcdefghijklmnopqrstuvwxyz|!$%&*+-./:<=>?@^_~")_
  private def isIdentComponentCharacter(c: Char): Boolean =
    isIdentStartCharacter(c) || isNumberCharacter(c)
  private def isStringComponentCharacter(c: Char): Boolean =
    !inCharSet("\n\r\"")(c)

  def readStringWhile(predicate: Char=>Boolean): String = {
    val buffer = new StringBuilder()
    while (nextCharAsInt != -1 && predicate(nextChar)) {
      buffer append nextChar
      skipChar()
    }
    buffer.toString()
  }
}
